from __future__ import annotations

# Std lib
from typing import TypedDict, Unpack, Tuple

# 3rd party lib
from shapely import Polygon, Point
from shapely.affinity import scale

# Loacal lib
from .shape import Shape


class CircleKwargs(TypedDict, total=False):
    x: float
    y: float
    radius: float
    radius_2: float
    rotation: float
    millable: bool


class Circle(Shape):

    # region Class Body

    # Class variables
    _default_geometry = Shape._default_geometry.copy()
    _default_geometry.update({"radius": 50, "radius_2": None})

    # Instance variables
    _radius: float
    _radius_2: float

    __slots__ = ["_radius", "_radius_2"]

    # endregion Class Body

    # region Dev Methods

    def __init__(self, **kwargs: Unpack[CircleKwargs]) -> None:
        super().__init__(**kwargs)

    def __repr__(self) -> str:
        radius_2 = f"\tradius 2: {self._radius_2}\n" if self._radius_2 is not None else ""
        output = (
            f"Shape: Circle\n"
            f"\tx: {self._x}\n"
            f"\ty: {self._y}\n"
            f"\tradius: {self._radius}\n"
            f"{radius_2}"
            f"\trotation (deg): {self._rotation}\n"
        )
        return output

    def _get_polygon(self) -> Polygon:

        radius_2 = self._radius if self._radius_2 is None else self._radius_2
        polygon = Point((0, 0)).buffer(1, resolution=100)
        polygon = self._apply_translation(polygon)
        polygon = scale(polygon, self._radius, radius_2)
        return self._apply_rotation(polygon)

    def _get_max(self) -> Tuple[float, float]:
        # Fetch the secondary radius
        if self._radius_2 is None:
            radius_2 = self._radius
        else:
            radius_2 = self._radius_2

        if self._rotation in [0, 180]:
            return self._x + self.radius, self._y + radius_2
        elif self._rotation in [90, 270]:
            return self._x + radius_2, self._y + self._radius
        else:
            return super()._get_max()

    def _get_min(self) -> Tuple[float, float]:
        # Fetch the secondary radius
        if self._radius_2 is None:
            radius_2 = self._radius
        else:
            radius_2 = self._radius_2

        if self._rotation in [0, 180]:
            return self._x - self.radius, self._y - radius_2
        elif self._rotation in [90, 270]:
            return self._x - radius_2, self._y - self._radius
        else:
            return super()._get_min()

    # endregion Dev Methods

    # region User Methods

    def copy(self, **kwargs) -> Circle:
        return super().copy(**kwargs)

    # endregion User Methods

    # region Properties

    @property
    def radius(self) -> float:
        return self._radius

    @radius.setter
    def radius(self, value: float):
        self._radius = float(value)

    @property
    def radius_2(self) -> float:
        if self._radius_2 is None:
            return self._radius
        return self._radius_2

    @radius_2.setter
    def radius_2(self, value: float):
        self._radius_2 = float(value)

    # endregion Properties


from .conversions import (length_units, current_units, voltage_units, time_units, dose_units,
                          LENGTHUNITS, CURRENTUNITS, VOLTAGEUNITS, TIMEUNITS, DOSEUNITS)
from .functions import plot_stream_file
from .enums import Material

__all__ = ["length_units", "current_units", "voltage_units", "LENGTHUNITS", "CURRENTUNITS", "VOLTAGEUNITS", "DOSEUNITS",
           "dose_units", "time_units", "TIMEUNITS", "plot_stream_file", "Material"]



from typing import Literal

# region Constants and Literals

LENGTHUNITS = Literal["m", "cm", "mm", "um", "nm"]
CURRENTUNITS = Literal["a", "ma", "ua", "na", "pa"]
DOSEUNITS = Literal["c", "mc", "uc", "nc", "pc"]
VOLTAGEUNITS = Literal["v", "kv"]
TIMEUNITS = Literal["s", "ms", "us", "ns"]
DECIMALS = 15
CONVERSIONS = {
    "m": 1,
    "cm": 100,
    "mm": 1000,
    "um": 1e6,
    "nm": 1e9,
    "a": 1,
    "ma": 1000,
    "ua": 1e6,
    "na": 1e9,
    "pa": 1e12,
    "c": 1,
    "mc": 1000,
    "uc": 1e6,
    "nc": 1e9,
    "pc": 1e12,
    "v": 1,
    "kv": 1000,
    "s": 1,
    "ms": 1000,
    "us": 1e6,
    "ns": 1e9
}


# endregion Constants and Literals

# region Functions

def convert_units(value: float, from_unit: str, to_unit: str, power: int = 1) -> float:
    from_factor = CONVERSIONS[from_unit]
    to_factor = CONVERSIONS[to_unit]
    converted_value = value * (to_factor / from_factor) ** power
    return round(converted_value, DECIMALS)


def dose_units(value: float, original: DOSEUNITS, to: DOSEUNITS, power: int = 1) -> float:
    return convert_units(value, from_unit=original, to_unit=to, power=power)


def length_units(value: float, original: LENGTHUNITS, to: LENGTHUNITS, power: int = 1) -> float:
    return convert_units(value, from_unit=original, to_unit=to, power=power)


def current_units(value: float, original: CURRENTUNITS, to: CURRENTUNITS, power: int = 1) -> float:
    return convert_units(value, from_unit=original, to_unit=to, power=power)


def voltage_units(value: float, original: VOLTAGEUNITS, to: VOLTAGEUNITS) -> float:
    return convert_units(value, from_unit=original, to_unit=to)


def time_units(value: float, original: TIMEUNITS, to: TIMEUNITS) -> float:
    return convert_units(value, original, to)

# endregion Functions



from __future__ import annotations

# Std lib
from typing import TypedDict, Unpack, Tuple

# 3rd party lib
from shapely import Polygon, Point, MultiPolygon, box, unary_union
from shapely.affinity import scale

# Loacal lib
from .shape import Shape


class RectangleKwargs(TypedDict, total=False):
    x: float
    y: float
    x_span: float
    y_span: float
    corner_radius: float
    rotation: float
    millable: bool


class Rectangle(Shape):

    # region Class Body

    # Class variables
    _default_geometry = Shape._default_geometry.copy()
    _default_geometry.update({"x_span": 100, "y_span": 100, "corner_radius": 0})

    # Instance variables
    _x_span: float
    _y_span: float
    _corner_radius: float

    __slots__ = ["_x_span", "_y_span", "_corner_radius"]

    # endregion Class Body

    # region Dev Methods

    def __init__(self, **kwargs: Unpack[RectangleKwargs]) -> None:
        super().__init__(**kwargs)
        if self._corner_radius * 2 > self._x_span or self._corner_radius * 2 > self._y_span:
            raise ValueError(f"Corner radius cannot be greater than x/y-span.")

    def __repr__(self) -> str:
        output = (
            f"Shape: Rectangle\n"
            f"\tx: {self._x}\n"
            f"\ty: {self._y}\n"
            f"\tx span: {self._x_span}\n"
            f"\ty span: {self._y_span}\n"
            f"\tcorner radius: {self._corner_radius}\n"
            f"\trotation (deg): {self._rotation}\n"
        )
        return output

    def _get_polygon(self) -> Polygon:

        polygon = scale(box(-0.5, -0.5, 0.5, 0.5), self._x_span, self._y_span)
        if self._corner_radius != 0:
            x, y = polygon.exterior.coords.xy
            radius = self._corner_radius
            points = [Point(min(x) + radius, max(y) - radius), Point(min(x) + radius, min(y) + radius),
                      Point(max(x) - radius, max(y) - radius), Point((max(x) - radius), min(y) + radius)]
            circles: MultiPolygon = unary_union([point.buffer(radius, resolution=100) for point in points])
            polygon = circles.convex_hull
        polygon = self._apply_translation(polygon)
        return self._apply_rotation(polygon)

    def _get_max(self) -> Tuple[float, float]:

        if self._rotation in [0, 180]:
            return self._x + self._x_span/2, self._y + self._y_span/2
        elif self._rotation in [90, 270]:
            return self._x + self._y_span/2, self._y + self._x_span/2
        else:
            return super()._get_max()

    def _get_min(self) -> Tuple[float, float]:

        if self._rotation in [0, 180]:
            return self._x - self._x_span/2, self._y - self._y_span/2
        elif self._rotation in [90, 270]:
            return self._x - self._y_span/2, self._y - self._x_span/2
        else:
            return super()._get_min()

    # endregion Dev Methods

    # region User Methods

    def copy(self, **kwargs) -> Rectangle:
        return super().copy(**kwargs)

    # endregion User Methods

    # region Properties

    @property
    def x_span(self) -> float:
        return self.x_span

    @x_span.setter
    def x_span(self, value: float):
        self.x_span = float(value)

    @property
    def y_span(self) -> float:
        return self._y_span

    @y_span.setter
    def y_span(self, value: float):
        self.y_span = float(value)

    @property
    def corner_radius(self) -> float:
        return self._corner_radius

    @corner_radius.setter
    def corner_radius(self, radius: float) -> None:
        if radius * 2 > self._x_span or radius * 2 > self._y_span:
            raise  ValueError(f"Corner corner radius * 2 cannot be greater than the spans.")
        self._corner_radius = radius

    # endregion Properties



from ..resources import time_units, TIMEUNITS, LENGTHUNITS, length_units


class Mill:

    # region Class Body

    _overlap: float
    _dwell_time: float
    _depth: float

    __slots__ = ["_overlap", "_dwell_time", "_depth"]

    # endregion Class Body

    # region Dev Methods

    def __init__(self) -> None:
        self.set_overlap(0.5)
        self.set_dwell_time(1, "us")
        self.set_mill_depth(0.5, "um")

    # endregion Dev Methods

    # region User Methods

    def get_mill_depth(self, units: LENGTHUNITS = "um") -> float:
        """Returns the milling depth in the desired units."""
        return length_units(self._depth, "m", units)

    def set_mill_depth(self, depth: float, units: LENGTHUNITS = "um"):
        """Sets the milling depth in the provided units."""
        self._depth = length_units(depth, units, "m")

    def get_overlap(self) -> float:
        """Returns the fraction of overlap between adjacent beam points in nanometers [nm]."""
        return self._overlap

    def set_overlap(self, overlap: float) -> None:
        """Sets the fraction of overlap between adjacent beam spots in nanometers [nm]."""
        if overlap >= 1:
            raise ValueError(f"The overlap fraction must be below 1, corresponding to 100%, not '{overlap}'.")
        self._overlap = overlap

    def get_dwell_time(self, units: TIMEUNITS = "us") -> float:
        """Returns the base dwell time pr. point in the desired time units."""
        return time_units(self._dwell_time, "s", units)

    def set_dwell_time(self, dwell_time: float, units: TIMEUNITS = "us") -> None:
        """Sets the base dwell time pr. point in the desired units."""
        dwell = time_units(dwell_time, units, "s")
        if dwell < 25e-9:
            raise ValueError(f"The minimum dwell time is 25 nanoseconds. Got '{dwell_time} {units}'.")
        self._dwell_time = dwell

    # endregion User Methods



from __future__ import annotations

# Std lib
from abc import ABC, abstractmethod
from typing import List, Unpack, Tuple, Literal, Callable, ClassVar, Type

# 3rd party lib
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib.patches import PathPatch, Circle
from matplotlib.path import Path
from numpy.typing import NDArray
from shapely import Polygon, MultiPolygon, GeometryCollection, unary_union

# Local lib
from .. import shapes
from ..beams import BeamType
from ..mill import Mill
from ..resources import TIMEUNITS, length_units, Material, LENGTHUNITS, time_units
from ..scan_styles.contour import contour

BOUNDARIES = Literal["x_min", "x_max", "y_min", "y_max"]


class FIBInterface(ABC):
    _fov_pixels: ClassVar[int]

    @abstractmethod
    def get_pixelsize(self, units: str) -> float:
        """Returns the pixelsize in the specified units for the current FIB settings."""
        ...

    @abstractmethod
    def get_fov(self, units: str) -> float:
        ...

    @property
    @abstractmethod
    def beam(self) -> BeamType:
        ...


class Add:
    # region Class Body

    _pattern: Pattern

    # endregion Class Body

    # region Dev Methods

    def __init__(self, pattern: Pattern):
        self._pattern = pattern

    # endregion Dev Methods

    # region User Methods

    def rectangle(self, **kwargs: Unpack[shapes.RectangleKwargs]) -> shapes.Rectangle:
        rect = shapes.Rectangle(pattern=self._pattern, **kwargs)  # type: ignore
        self._pattern._add_shape(rect)
        return rect

    def circle(self, **kwargs: Unpack[shapes.CircleKwargs]) -> shapes.Circle:
        circ = shapes.Circle(pattern=self._pattern, **kwargs)  # type: ignore
        self._pattern._add_shape(circ)
        return circ

    def polygon(self, **kwargs: Unpack[shapes.PolyKwargs]) -> shapes.Poly:
        poly = shapes.Poly(pattern=self._pattern, **kwargs)  # type: ignore
        self._pattern._add_shape(poly)
        return poly

    def regular_polygon(self, **kwargs: Unpack[shapes.RegularPolygonKwargs]) -> shapes.RegularPolygon:
        poly = shapes.RegularPolygon(pattern=self._pattern, **kwargs)  # type: ignore
        self._pattern._add_shape(poly)
        return poly

    def lattice(self, **kwargs: Unpack[shapes.LatticeKwargs]) -> shapes.Lattice:
        lattice = shapes.Lattice(pattern=self._pattern, **kwargs)  # type: ignore
        self._pattern._add_shape(lattice)
        return lattice

    # endregion User Methods


class Pattern:

    # region Class Body

    _shapes: List[shapes.Shape]
    _mill: Mill
    _fib: FIBInterface
    _x: float
    _y: float
    _pass_multiplier: int
    _scan_style: Callable
    _material: Material
    Materials: Type[Material]
    add: Add

    __slots__ = ["_shapes", "_mill", "_fib", "_x", "_y", "_pass_multiplier", "add", "_scan_style", "_material",
                 "Materials"]

    # endregion Class Body

    # region Dev Methods

    def __init__(self, fib, x: float = 0, y: float = 0) -> None:
        self._fib = fib
        self._shapes = []
        self._mill = Mill()
        self.add = Add(self)
        self.set_position(x, y)
        self._scan_style = contour
        self._material = Material.Default_Si
        self.Materials = Material

    def _add_shape(self, shape: shapes.Shape) -> None:
        self._shapes.append(shape)

    def _get_millable_area(self) -> Polygon | MultiPolygon | GeometryCollection:

        millable = []
        non_millable = []

        def filter_millable(shape_to_filter: shapes.Shape) -> None:
            """Recursively filter all shapes based on the milling flag."""
            if isinstance(shape_to_filter, shapes.CombinedShape):
                for shape in shape_to_filter._shapes:
                    filter_millable(shape)
            elif isinstance(shape_to_filter, shapes.Lattice):
                for row in shape_to_filter._shapes:
                    for shape in row:
                        if shape is not None:
                            filter_millable(shape)
            elif isinstance(shape_to_filter, shapes.Shape):
                if shape_to_filter._millable:
                    millable.append(shape_to_filter.polygon)
                else:
                    non_millable.append(shape_to_filter.polygon)
            else:
                raise TypeError(f"Got '{type(shape_to_filter)}', expected 'Shape' or any of it's subclasses.")
            return

        for pattern_shape in self._shapes:
            filter_millable(pattern_shape)

        millable_area = unary_union(millable)
        non_millable_area = unary_union(non_millable)
        millable_area = millable_area.difference(non_millable_area)

        return millable_area

    def _get_pattern_ax(self, ax: plt.Axes = None, alpha: float = 1) -> plt.Axes:
        """Makes a plot where the millable area is filled in, not including any area containing a non-millable shape."""

        millable_area = self._get_millable_area()

        if ax is None:
            ax = plt.gca()

        if millable_area.is_empty:
            print("No millable area to plot.")
        else:
            if millable_area.geom_type == "MultiPolygon":
                for poly in millable_area.geoms:
                    exterior_coords = np.array(poly.exterior.coords)
                    interiors = [np.array(interior.coords) for interior in poly.interiors]
                    path = Path.make_compound_path(Path(exterior_coords), *[Path(interior) for interior in interiors])
                    patch = PathPatch(path, facecolor='black', edgecolor='none', alpha=alpha, label="Millable")
                    ax.add_patch(patch)
            else:
                poly = millable_area
                exterior_coords = np.array(poly.exterior.coords)
                interiors = [np.array(interior.coords) for interior in poly.interiors]
                path = Path.make_compound_path(Path(exterior_coords), *[Path(interior) for interior in interiors])
                patch = PathPatch(path, facecolor='black', edgecolor='none', alpha=alpha, label="Millable")
                ax.add_patch(patch)

        return ax

    def _pixelate_beam_path(self, streamlines: NDArray) -> NDArray:

        # Conversion factor: nm -> pixels
        conversion_factor = 1 / self._fib.get_pixelsize("nm")

        # Extract physical positions (assuming column 1 = x, column 2 = y)
        streamlines[:, 1] = np.rint(streamlines[:, 1] * conversion_factor)
        streamlines[:, 2] = np.rint(streamlines[:, 2] * conversion_factor)

        return streamlines

    def _get_passes(self, streamlines: NDArray = None) -> int:

        millable_area = length_units(self._get_millable_area().area, "nm", "um", power=2)
        millable_volume = millable_area * self._mill.get_mill_depth("um")
        volume_pr_dose = self._material.value
        time_goal = millable_volume / (self._fib.beam.get_current("na") * volume_pr_dose)

        effective_mill_time = 0
        for vec in streamlines:
            if vec.shape[0] == 4:
                if vec[3] == 0:
                    continue
            effective_mill_time += vec[0]
        effective_mill_time = time_units(effective_mill_time * 100, "ns", "s")

        nr_passes = np.rint(time_goal/effective_mill_time).astype(int)

        return nr_passes

    # endregion Dev Methods

    # region User Methods

    def set_material(self, material: Material) -> None:
        """Sets the material the sample is made out of. Materials can be found under the Materials variable of the
        pattern instance."""
        self._material = material

    def write_stream(self, filename: str) -> None:
        streamlines = np.vstack(self._scan_style(self._get_millable_area(), self._mill, self._fib.beam, rasterize=True))
        streamlines = self._pixelate_beam_path(streamlines)
        passes = self._get_passes(streamlines)

        streamlines[:, 1:3] += np.array([self._fib._fov_pixels // 2, self._fib._fov_pixels // 2])
        if not filename.endswith(".str"):
            filename += ".str"
        with open(filename, "w", encoding="utf-8") as file:
            file.writelines(["s16\n", str(streamlines.shape[0]) + "\n", str(passes) + "\n"])
            file.writelines([f"{int(line[0])} {int(line[1])} {int(line[2])}"
                             f"{' ' + str(int(line[3])) + "\n" if line[3] != 2 else '\n'}"
                             for line in streamlines])
        return

    def set_position(self, x: float = 0, y: float = 0) -> None:
        """Sets the position of the pattern, including all contained shapes."""
        self._x, self._y = x, y

    def plot_pattern(self, figsize: Tuple[float, float] = (15, 10), alpha: float = 0.6) -> None:
        """Makes a plot where the millable area is filled in, not including any area containing a non-millable shape."""
        plt.rcParams["figure.figsize"] = figsize
        prev_pos = self._x, self._y
        self.set_position(0, 0)
        ax = self._get_pattern_ax(None, alpha)
        self.set_position(*prev_pos)
        ax.relim()
        ax.autoscale_view()
        ax.set_aspect('equal')
        # Retrieve the current handles and labels
        handles, labels = plt.gca().get_legend_handles_labels()

        # Create a dictionary to remove duplicates. In Python 3.7+, dict preserves insertion order.
        by_label = dict(zip(labels, handles))

        # Now pass the unique handles and labels to plt.legend()
        plt.legend(by_label.values(), by_label.keys())
        plt.show()

    def plot_beam_path(self, figsize: Tuple[float, float] = (15, 10), alpha: float = 0.25, rasterize: bool = True) -> None:

        millable_area = self._get_millable_area()

        lines = self._scan_style(millable_area, self._mill, self._fib.beam, rasterize=rasterize)

        plt.figure(figsize=figsize)
        ax = self._get_pattern_ax(None, alpha)
        for line in lines:
            plt.plot(line[:, 1], line[:, 2], "o-")

        plt.gca().set_aspect("equal")
        plt.show()

    def animate_beam_path(self, figsize=(15, 10), alpha=0.25, repeat: bool = False):
        """
        Animates drawing the beam path contours with fixed axis limits.
        The axis limits are computed from all contour line data so that the
        plot’s size remains constant throughout the animation.
        """

        # Get the overall geometry and compute the contour lines.
        millable_area = self._get_millable_area()
        lines = contour(millable_area, self._mill, self._fib.beam, rasterize=True)

        # Precompute coordinate data for each line.
        # Each entry in line_data is a tuple (x_values, y_values)
        line_data = []
        for line in lines:
            x, y = line[:, 1], line[:, 2]
            line_data.append((list(x), list(y)))

        # Compute the overall bounding box from all lines.
        all_x = [x for (x_vals, _) in line_data for x in x_vals]
        all_y = [y for (_, y_vals) in line_data for y in y_vals]
        margin = 1  # optional extra space around the data
        x_min, x_max = min(all_x) - margin, max(all_x) + margin
        y_min, y_max = min(all_y) - margin, max(all_y) + margin

        # Create the figure and axis with fixed limits.
        fig, ax = plt.subplots(figsize=figsize)
        ax.set_xlim(x_min, x_max)
        ax.set_ylim(y_min, y_max)
        ax.set_aspect("equal")

        # State dictionary to track animation progress.
        state = {
            'current_line': 0,  # index of the line being drawn
            'current_point': 0,  # index of the next point in the current line
            'drawn_lines': []  # list of completed lines
        }

        # Determine the total number of frames (one per point).
        total_frames = sum(len(x_vals) for (x_vals, _) in line_data)

        beam_radius = self._fib.beam.get_radius("nm")

        def update(frame):
            """
            Update function for FuncAnimation.
            Adds one point to the current line at each frame and then moves on.
            """
            # Update the state: draw one more point on the current line.
            if state['current_line'] < len(line_data):
                x_vals, y_vals = line_data[state['current_line']]
                if state['current_point'] < len(x_vals):
                    state['current_point'] += 1
                else:
                    # Current line is finished; store it and move to next line.
                    state['drawn_lines'].append((x_vals, y_vals))
                    state['current_line'] += 1
                    state['current_point'] = 0

            # Clear the axis and reapply fixed limits.
            ax.clear()
            ax.set_xlim(x_min, x_max)
            ax.set_ylim(y_min, y_max)
            ax.set_aspect("equal")

            # Optionally, reapply any background or styling here (e.g., via self._get_pattern_ax).

            # Draw all completed lines (blue).
            for x_vals, y_vals in state['drawn_lines']:
                ax.plot(x_vals, y_vals, alpha=0.3)
            # Draw the current line (partial, in red) if it exists.
            if state['current_line'] < len(line_data):
                x_vals, y_vals = line_data[state['current_line']]
                ax.plot(x_vals[:state['current_point']], y_vals[:state['current_point']], color='red')

                last_x = x_vals[state['current_point'] - 1]
                last_y = y_vals[state['current_point'] - 1]

                # Add a circle with radius 6.5 at the last plotted point
                circle = Circle((last_x, last_y), beam_radius, color='black',
                                fill=True, alpha=0.5, linewidth=0)
                ax.add_patch(circle)

            return []  # When not using blit, returning an empty list is fine.

        # Create the animation.
        anim = FuncAnimation(fig, update, frames=total_frames, interval=10, blit=False, repeat=repeat,
                             cache_frame_data=True)

        plt.show()

    def get_dwell_time(self, units: TIMEUNITS) -> float:
        return self._mill.get_dwell_time(units)

    def set_dwell_time(self, dt: float, units: TIMEUNITS) -> None:
        self._mill.set_dwell_time(dt, units)

    def set_depth(self, depth: float, units: LENGTHUNITS = "um") -> None:
        self._mill.set_mill_depth(depth, units)

    def get_depth(self, units: LENGTHUNITS) -> float:
        return self._mill.get_mill_depth(units)

    def get_overlap(self) -> float:
        return self._mill.get_overlap()

    def set_overlap(self, overlap_fraction: float) -> None:
        self._mill.set_overlap(overlap_fraction)

    def set_pass_multiplier(self, multiplier: int) -> None:
        """ Sets the number of hardcoded passes for the beam path.

            - If set to 1, the beam will trace the pattern once in the exported stream file.
              The number of actual passes performed will then be determined by the "nr. of passes" parameter in the
              stream file or an override in the FIB machine's interface.
            - If set to ie. 2, the beam will trace the pattern twice in the exported stream file.
              This is equivalent to setting the multiplier to 1 and adjusting the "nr. of passes" parameter to 2.

            This feature is particularly useful for parameter sweeps. For example, suppose you need to create two
            versions of the same pattern with different dwell times while using the same stream file.
            To achieve the same milling depth, the number of passes must vary. However, since the stream file
            enforces a single pass setting for the entire file, you can work around this by adjusting
            the pass multiplier.

            For instance, if one pattern has a 1 µs dwell time and another has 0.5 µs, setting the pass multiplier of
            the 0.5 µs pattern to 2 ensures both receive the same dose.
        """
        # assert isinstance(multiplier, int)
        # self._pass_multiplier = multiplier
        raise NotImplementedError(f"This feature is currently not supported. Will be in later versions.")

    def place_next_to(self, pattern: Pattern, boundary: BOUNDARIES, offset: float = 0) -> None:
        """Places the pattern adjacent to the specified boundary of another pattern."""
        direction = -1 if "min" in boundary else 1
        pos = getattr(pattern, boundary) + (getattr(self, f"{boundary[0]}_span") / 2) * direction + offset
        self.set_position(**{boundary[0]: pos})

    # endregion User Methods

    # region Properties

    @property
    def x(self) -> float:
        return self._x

    @x.setter
    def x(self, x: float) -> None:
        self.set_position(x=x)

    @property
    def y(self) -> float:
        return self._y

    @y.setter
    def y(self, y: float) -> None:
        self.set_position(y=y)

    @property
    def x_span(self) -> float:
        return self.x_max - self.x_min

    @property
    def y_span(self) -> float:
        return self.y_max - self.y_min

    @property
    def x_min(self) -> float:
        return min([shape.x_min for shape in self._shapes]) + self._x

    @property
    def x_max(self) -> float:
        return max([shape.x_max for shape in self._shapes]) + self._x

    @property
    def y_min(self) -> float:
        return min([shape.y_min for shape in self._shapes]) + self._y

    @property
    def y_max(self) -> float:
        return max([shape.y_max for shape in self._shapes]) + self._y

    # endregion Properties



from __future__ import annotations

# Std lib
from typing import TypedDict, Unpack, Tuple, Iterable, List

# 3rd party lib
from shapely import Polygon
from shapely.affinity import translate

# Local lib
from .shape import Shape


class PolyKwargs(TypedDict, total=False):
    millable: bool
    x: float
    y: float
    vertices: Iterable[Tuple[float, float]]
    rotation: float


class Poly(Shape):

    # region Class Body

    _default_geometry = Shape._default_geometry.copy()
    _default_geometry.update({"vertices": [(3.061616997868383e-15, 50.0), (-43.30127018922193, -25.000000000000007),
                                           (43.30127018922192, -25.00000000000002)]})

    _vertices: List[Tuple[float, float]]

    # endregion Class Body

    # region Dev Methods

    def __init__(self, **kwargs: Unpack[PolyKwargs]):
        super().__init__(**kwargs)

    def _get_polygon(self, **kwargs: Unpack[PolyKwargs]) -> Polygon:

        polygon = Polygon(self._vertices)
        polygon = translate(polygon, -polygon.centroid.x, -polygon.centroid.y)

        # Apply translation and rotation if needed
        polygon = self._apply_translation(polygon)
        polygon = self._apply_rotation(polygon)

        # Initialize the shape with final polygon
        return polygon

    # endregion Dev Methods

    # region User Methods

    def copy(self, **kwargs: Unpack[PolyKwargs]) -> Poly:
        return super().copy(**kwargs)

    # endregion User Methods

    # region Properties

    @property
    def vertices(self) -> List[Tuple[float, float]]:
        """Returns a list of the polygon's vertices."""
        return self._vertices

    @vertices.setter
    def vertices(self, vertices: List[Tuple[float, float]]) -> None:
        """Sets the list of vertices for the polygon."""
        self._vertices = vertices

    # endregion Properties



from __future__ import annotations

# Std lib
import gc
from typing import List, TypedDict, Unpack, Tuple
from copy import copy

# 3rd party lib
import numpy as np
from numpy import array, deg2rad, vstack, column_stack
from numpy.typing import NDArray
from shapely import Polygon, MultiPolygon, unary_union

# Local lib
from .shape import Shape


class LatticeKwargs(TypedDict, total=False):
    x: float
    y: float
    rows: int
    cols: int
    alpha: float
    beta: float
    gamma: float
    rotation: float
    shape: Shape


class CopyKwargs(TypedDict, total=False):
    x: float
    y: float


class Lattice(Shape):

    # region Class Body

    # Class variables
    _default_geometry = Shape._default_geometry.copy()
    _default_geometry.update({"rows": 2, "cols": 2, "alpha": 200, "beta": 200, "gamma": 90})

    # Instance variables
    _sites: NDArray
    _rows: int
    _cols: int
    _alpha: float
    _beta: float
    _gamma: float
    _shapes: List[List[Shape | None]]

    # endregion Class Body

    # region Dev Methods

    def __init__(self, **kwargs: Unpack[LatticeKwargs]) -> None:
        shape = kwargs.pop("shape", None)
        super().__init__(**kwargs)

        self._construct_site_matrix()
        if shape:
            self._create_shapes(shape)

    def _construct_first_unit_cell(self) -> NDArray:
        """
        Create the first unit cell vertices using the unit vectors.
        Returns two rows of vertices:
          - bottom_row: [bottom_left, bottom_right]
          - top_row:    [top_left, top_right]
        Also returns the unit vectors v1 and v2.
        """
        gamma_rad = deg2rad(self._gamma)
        v1 = np.array([self._alpha, 0])
        v2 = np.array([self._beta * np.cos(gamma_rad), self._beta * np.sin(gamma_rad)])

        bottom_left = np.array([0, 0])
        bottom_right = v1
        top_left = v2
        top_right = v1 + v2

        bottom_row = np.array([bottom_left, bottom_right])
        top_row = np.array([top_left, top_right])

        return array([bottom_row, top_row])

    def _construct_site_matrix(self) -> None:
        """Constructs a 2D matrix that contains the coordinates for all lattice sites. Deletes all previous shapes
        and re-initializes the list containg the lattice's shapes where each element is None, allowing assignment
        through indexing later."""

        unit_cell = self._construct_first_unit_cell()
        lattice = unit_cell.copy()

        if self._rows == 1:
            lattice = lattice[0, :]
        elif self._rows == 2:
            pass
        else:
            for i in range(self._rows - 2):
                this_row = lattice[-(2), :].copy() + array([0, unit_cell[1, 0, 1] * 2])
                lattice = vstack((lattice, array([this_row])))

        if self._cols == 1:
            lattice = lattice[:, 0]
        elif self._cols == 2:
            pass
        else:
            for _ in range(self._cols - 2):
                this_col = array([lattice[:, -2].copy() + array([unit_cell[0, 1, 0] * 2, 0])])
                this_col = this_col.transpose(1, 0, 2)
                lattice = column_stack((lattice, array([this_col])[0]))

        lattice -= array([np.max(lattice[:, :, 0]) / 2, np.max(lattice[:, :, 1]) / 2])

        # Delete all shapes contained in the lattice
        if hasattr(self, "_shapes"):
            for row in self._shapes:
                for col in row:
                    del col  # Remove references to objects

            self._shapes.clear()  # Remove references from the list itself
            gc.collect()  # Ensure immediate cleanup

        # Re-initialize indexable list of shapes
        self._shapes = [[None for i in range(lattice.shape[1])] for j in range(lattice.shape[0])]  # type: ignore

        # Assign site matrix to member variable
        self._sites = lattice

    def _create_shapes(self, shape: Shape, rows: Tuple[int, int] = None, cols: Tuple[int, int] = None) -> None:
        """Create a shape at each lattice site. If previously occupied, delete and replace."""
        if rows is None:
            rows = (0, self._sites.shape[0])
        if cols is None:
            cols = (0, self._sites.shape[1])

        # Set the original shape at the first position
        shape.set_position(*self._sites[0, 0])
        skip = True
        for ridx, row in enumerate(self._sites[rows[0]:rows[1]]):
            for cidx, col in enumerate(row[cols[0]:cols[1]]):
                del self._shapes[ridx][cidx]
                if skip:
                    self._shapes[ridx].insert(cidx, shape)
                    skip = False
                    continue
                self._shapes[ridx].insert(cidx, shape.copy(x=col[0], y=col[1]))

    def _get_polygon(self) -> Polygon | MultiPolygon:
        shapes = []
        for row in self._shapes:
            for s in row:
                if s is not None:
                    shapes.append(s.polygon)
        return unary_union(shapes)

    def __getitem__(self, key) -> Lattice:
        """
        Enable indexing and slicing for the Lattice.

        If a two-dimensional slice is used (e.g., lattice[1:3, 2:4]),
        return a new Lattice object containing the corresponding subset of sites.
        If two integer indices are provided (e.g., lattice[i, j]),
        return the site object at that position.
        If only one index is provided, treat it as row indexing and return that row as a list.
        """
        if isinstance(key, tuple):
            if len(key) != 2:
                raise IndexError("Lattice indexing supports two indices (row, col)")

            row_idx, col_idx = key

            # Handle slices properly
            if isinstance(row_idx, slice) or isinstance(col_idx, slice):
                row_slice = self._shapes[row_idx]  # Select relevant rows

                # Ensure row_slice is always a list of lists
                if isinstance(row_slice, list) and row_slice and not isinstance(row_slice[0], list):
                    row_slice = [row_slice]

                    # Get the subset of sites and return as a new Lattice object
                new_sites = self._sites[row_idx, col_idx]
                if isinstance(row_idx, int) or isinstance(col_idx, int):
                    if isinstance(col_idx, int):
                        shapes = [[row[col_idx]] for row in row_slice]
                    else:
                        shapes = [row[col_idx] for row in row_slice]
                    min_x, max_x = np.min(new_sites[:, 0]), np.max(new_sites[:, 0])
                    min_y, max_y = np.min(new_sites[:, 1]), np.max(new_sites[:, 1])
                else:
                    shapes = [row[col_idx] for row in row_slice]
                    min_x, max_x = np.min(new_sites[:, :, 0]), np.max(new_sites[:, :, 0])
                    min_y, max_y = np.min(new_sites[:, :, 1]), np.max(new_sites[:, :, 1])
                new_pos = (min_x + (max_x - min_x) / 2, min_y + (max_y - min_y) / 2)

            # Single element access
            else:
                shapes = [[self._shapes[row_idx][col_idx]]]
                new_sites = array([[self._sites[row_idx, col_idx]]])
                new_pos = new_sites[0, 0, 0], new_sites[0, 0, 1]

        # If a single integer is provided, return the full row
        elif isinstance(key, int):
            shapes = self._shapes[key]
            new_sites = self._sites[key]
            x_min = np.min(new_sites[:, 0])
            x_max = np.max(new_sites[:, 0])
            new_pos = (x_min + (x_max - x_min) / 2, new_sites[0, 0])
        else:
            raise IndexError(f"Invalid index. Got '{key}'.")

        new_rows, new_cols = new_sites.shape[:2]
        new_lattice = Lattice(rows=2, cols=2)
        new_lattice._rows = new_rows
        new_lattice._cols = new_cols
        new_lattice._pattern = self._pattern
        new_lattice._x, new_lattice._y = new_pos
        new_lattice._sites = new_sites
        new_lattice._shapes = shapes
        return new_lattice

    def __setitem__(self, key, shape: Shape | None):
        """
        Replace shapes in the lattice.

        If a single index (row, col) is provided, replace the shape at that position.
        If a slice is used, the first position is assigned `shape`, and the rest are filled with `shape.copy()`.
        """
        if not isinstance(shape, (Shape, type(None))):  # type: ignore
            raise TypeError(f"Invalid type. Expected type(Shape) or NoneType, got '{type(shape)}'.")
        elif isinstance(shape, Lattice):
            raise TypeError(f"Invalid type. Got '{type(shape)}'. "
                            f"Lattices cannot currently be assigned to points in another lattice.")

        if isinstance(key, tuple):
            if len(key) != 2:
                raise IndexError("Lattice indexing supports two indices (row, col)")

            row_idx, col_idx = key

            # Handle slices properly
            if isinstance(row_idx, slice) or isinstance(col_idx, slice):
                row_slice = self._shapes[row_idx]  # Select relevant rows

                # Ensure row_slice is always a list of lists
                if isinstance(row_slice, list) and row_slice and not isinstance(row_slice[0], list):
                    row_slice = [row_slice]

                # Apply shape and its copies
                first_set = False
                for i, row in enumerate(self._shapes[row_idx]):
                    for j in range(len(row[col_idx])):
                        # Delete the existing shape
                        delshape = row[col_idx][j]
                        row[col_idx].remove(delshape)
                        del delshape

                        # Assign shape to the first spot, then copy for others
                        if not first_set:
                            row[col_idx].insert(j, shape)
                            first_set = True
                        else:
                            if shape is not None:
                                row[col_idx].insert(j, shape.copy())
                            else:
                                row[col_idx].insert(j, shape)

            else:
                # Delete and replace a single shape
                delshape = self._shapes[row_idx][col_idx]
                self._shapes[row_idx][col_idx] = shape
                del delshape

        else:
            raise IndexError("Invalid index type. Use (row, col) indexing.")


    # endregion Dev Methods

    # region User Methods

    def rotate(self, rot_angle: float, rot_point: Tuple[float, float] | Shape = None) -> rotate:

        # Create the rotation matrix
        theta = np.deg2rad(rot_angle)
        self._sites = np.array([
            [np.cos(theta), -np.sin(theta)],
            [np.sin(theta), np.cos(theta)]
        ])

        for row, x in zip(self._shapes, self._sites):
            for shape, pos in zip(row, x):
                shape.rotate(rot_angle)
                shape.set_position(*pos)

    def set_position(self, x: float | Shape = None, y: float | Shape = None) -> None:

        old_pos = array([self._x, self._y])

        if x is not None:
            self._x = x
        else:
            x = 0
        if y is not None:
            self._y = y
        else:
            y = 0

        self._sites -= old_pos
        self._sites += array([x, y])

        for row, x in zip(self._shapes, self._sites):
            for shape, pos in zip(row, x):
                shape.set_position(*pos)

    def copy(self, **kwargs) -> Lattice:
        new_lattice = copy(self)
        new_lattice._shapes = [[shape.copy() for shape in row] for row in self._shapes]
        super().copy(**kwargs)
        return new_lattice

    # endregion User Methods

    # region Properties

    @property
    def x(self) -> float:
        """Getter for x coordinates"""
        return self._x

    @x.setter
    def x(self, new_x):
        """Setter for x coordinates. Updates site positions accordingly."""
        self.set_position(x=new_x)

    @property
    def y(self):
        """Getter for y coordinates"""
        return self._y

    @y.setter
    def y(self, new_y):
        """Setter for y coordinates. Updates site positions accordingly."""
        self.set_position(y=new_y)

    # endregion Properties



# Local lib
from .beam import Beam


class IonBeams:

    # region 500V
    class Acceleration500V:
        class pA1_5(Beam):
            _beam_diameter: float = 2.52e-07
            _current: float = 1.5e-12
            _voltage: float = 500

        class pA6(Beam):
            _beam_diameter: float = 4.98e-07
            _current: float = 6e-12
            _voltage: float = 500

        class pA15(Beam):
            _beam_diameter: float = 7.79e-07
            _current: float = 1.5e-11
            _voltage: float = 500

        class pA24(Beam):
            _beam_diameter: float = 9.96e-07
            _current: float = 2.4e-11
            _voltage: float = 500

        class pA42(Beam):
            _beam_diameter: float = 1.311e-06
            _current: float = 4.2e-11
            _voltage: float = 500

        class nA0_11(Beam):
            _beam_diameter: float = 2.136e-06
            _current: float = 1.1e-10
            _voltage: float = 500

        class nA0_18(Beam):
            _beam_diameter: float = 2.771e-06
            _current: float = 1.8e-10
            _voltage: float = 500

        class nA0_33(Beam):
            _beam_diameter: float = 3.96e-06
            _current: float = 3.3e-10
            _voltage: float = 500

        class nA0_92(Beam):
            _beam_diameter: float = 9.27e-06
            _current: float = 9.2e-10
            _voltage: float = 500

        class nA2(Beam):
            _beam_diameter: float = 2.442e-05
            _current: float = 2e-09
            _voltage: float = 500

        class nA2_8(Beam):
            _beam_diameter: float = 3.8445e-05
            _current: float = 2.8e-09
            _voltage: float = 500

        class nA21(Beam):
            _beam_diameter: float = 0.00075
            _current: float = 2.1e-08
            _voltage: float = 500

    # endregion 500V

    # region 1kV
    class Acceleration1kV:
        class pA0_7(Beam):
            _beam_diameter: float = 8e-08
            _current: float = 7e-13
            _voltage: float = 1

        class pA2_3(Beam):
            _beam_diameter: float = 1.1e-07
            _current: float = 2.3e-12
            _voltage: float = 1

        class pA5_3(Beam):
            _beam_diameter: float = 1.43e-07
            _current: float = 5.3e-12
            _voltage: float = 1

        class pA8_6(Beam):
            _beam_diameter: float = 1.68e-07
            _current: float = 8.6e-12
            _voltage: float = 1

        class pA14(Beam):
            _beam_diameter: float = 2e-07
            _current: float = 1.4e-11
            _voltage: float = 1

        class pA36(Beam):
            _beam_diameter: float = 2.88e-07
            _current: float = 3.6e-11
            _voltage: float = 1

        class pA59(Beam):
            _beam_diameter: float = 3.53e-07
            _current: float = 5.9e-11
            _voltage: float = 1

        class nA0_11(Beam):
            _beam_diameter: float = 4.59e-07
            _current: float = 1.1e-10
            _voltage: float = 1

        class nA0_3(Beam):
            _beam_diameter: float = 7.49e-07
            _current: float = 3e-10
            _voltage: float = 1

        class nA0_66(Beam):
            _beam_diameter: float = 1.092e-06
            _current: float = 6.6e-10
            _voltage: float = 1

        class nA0_92(Beam):
            _beam_diameter: float = 1.284e-06
            _current: float = 9.2e-10
            _voltage: float = 1

        class nA6_9(Beam):
            _beam_diameter: float = 3.927e-06
            _current: float = 6.9e-09
            _voltage: float = 1

    # endregion 1kV

    # region 2kV
    class Acceleration2kV:
        class pA0_3(Beam):
            _beam_diameter: float = 3.9e-08
            _current: float = 3e-13
            _voltage: float = 2

        class pA1_3(Beam):
            _beam_diameter: float = 5.1e-08
            _current: float = 1.3e-12
            _voltage: float = 2

        class pA3_2(Beam):
            _beam_diameter: float = 6.5e-08
            _current: float = 3.2e-12
            _voltage: float = 2

        class pA5_5(Beam):
            _beam_diameter: float = 7.7e-08
            _current: float = 5.5e-12
            _voltage: float = 2

        class pA10(Beam):
            _beam_diameter: float = 9.2e-08
            _current: float = 1e-11
            _voltage: float = 2

        class pA28(Beam):
            _beam_diameter: float = 1.26e-07
            _current: float = 2.8e-11
            _voltage: float = 2

        class pA46(Beam):
            _beam_diameter: float = 1.5e-07
            _current: float = 4.6e-11
            _voltage: float = 2

        class pA89(Beam):
            _beam_diameter: float = 1.95e-07
            _current: float = 8.9e-11
            _voltage: float = 2

        class nA0_25(Beam):
            _beam_diameter: float = 2.99e-07
            _current: float = 2.5e-10
            _voltage: float = 2

        class nA0_56(Beam):
            _beam_diameter: float = 4.32e-07
            _current: float = 5.6e-10
            _voltage: float = 2

        class nA0_77(Beam):
            _beam_diameter: float = 5.07e-07
            _current: float = 7.7e-10
            _voltage: float = 2

        class nA6_7(Beam):
            _beam_diameter: float = 1.935e-06
            _current: float = 6.7e-09
            _voltage: float = 2

    # endregion 2kV

    # region 5kV
    class Acceleration5kV:
        class pA0_4(Beam):
            _beam_diameter: float = 2e-08
            _current: float = 4e-13
            _voltage: float = 5

        class pA1_6(Beam):
            _beam_diameter: float = 2.7e-08
            _current: float = 1.6e-12
            _voltage: float = 5

        class pA4_4(Beam):
            _beam_diameter: float = 3.5e-08
            _current: float = 4.4e-12
            _voltage: float = 5

        class pA8(Beam):
            _beam_diameter: float = 4.1e-08
            _current: float = 8e-12
            _voltage: float = 5

        class pA16(Beam):
            _beam_diameter: float = 4.8e-08
            _current: float = 1.6e-11
            _voltage: float = 5

        class pA47(Beam):
            _beam_diameter: float = 6.7e-08
            _current: float = 4.7e-11
            _voltage: float = 5

        class pA81(Beam):
            _beam_diameter: float = 8e-08
            _current: float = 8.1e-11
            _voltage: float = 5

        class nA0_15(Beam):
            _beam_diameter: float = 9.6e-08
            _current: float = 1.5e-10
            _voltage: float = 5

        class nA0_45(Beam):
            _beam_diameter: float = 1.43e-07
            _current: float = 4.5e-10
            _voltage: float = 5

        class nA1(Beam):
            _beam_diameter: float = 2.07e-07
            _current: float = 1e-09
            _voltage: float = 5

        class nA1_5(Beam):
            _beam_diameter: float = 2.51e-07
            _current: float = 1.5e-09
            _voltage: float = 5

        class nA12(Beam):
            _beam_diameter: float = 1.4e-06
            _current: float = 1.2e-08
            _voltage: float = 5

    # endregion 5kV

    # region 8kV
    class Acceleration8kV:
        class pA0_5(Beam):
            _beam_diameter: float = 1.2e-08
            _current: float = 5e-13
            _voltage: float = 8

        class pA1_9(Beam):
            _beam_diameter: float = 1.6e-08
            _current: float = 1.9e-12
            _voltage: float = 8

        class pA5_7(Beam):
            _beam_diameter: float = 2.2e-08
            _current: float = 5.7e-12
            _voltage: float = 8

        class pA11(Beam):
            _beam_diameter: float = 2.6e-08
            _current: float = 1.1e-11
            _voltage: float = 8

        class pA21(Beam):
            _beam_diameter: float = 3e-08
            _current: float = 2.1e-11
            _voltage: float = 8

        class pA62(Beam):
            _beam_diameter: float = 4.1e-08
            _current: float = 6.2e-11
            _voltage: float = 8

        class nA0_11(Beam):
            _beam_diameter: float = 5e-08
            _current: float = 1.1e-10
            _voltage: float = 8

        class nA0_22(Beam):
            _beam_diameter: float = 6.3e-08
            _current: float = 2.2e-10
            _voltage: float = 8

        class nA0_66(Beam):
            _beam_diameter: float = 9e-08
            _current: float = 6.6e-10
            _voltage: float = 8

        class nA1_6(Beam):
            _beam_diameter: float = 1.33e-07
            _current: float = 1.6e-09
            _voltage: float = 8

        class nA2_2(Beam):
            _beam_diameter: float = 1.59e-07
            _current: float = 2.2e-09
            _voltage: float = 8

        class nA19(Beam):
            _beam_diameter: float = 1.056e-06
            _current: float = 1.9e-08
            _voltage: float = 8

    # endregion 8kV

    # region 16kV
    class Acceleration16kV:
        class pA0_6(Beam):
            _beam_diameter: float = 1.2e-08
            _current: float = 6e-13
            _voltage: float = 16

        class pA3_4(Beam):
            _beam_diameter: float = 1.9e-08
            _current: float = 3.4e-12
            _voltage: float = 16

        class pA11(Beam):
            _beam_diameter: float = 2.5e-08
            _current: float = 1.1e-11
            _voltage: float = 16

        class pA21(Beam):
            _beam_diameter: float = 2.9e-08
            _current: float = 2.1e-11
            _voltage: float = 16

        class pA45(Beam):
            _beam_diameter: float = 3.4e-08
            _current: float = 4.5e-11
            _voltage: float = 16

        class nA0_13(Beam):
            _beam_diameter: float = 4.5e-08
            _current: float = 1.3e-10
            _voltage: float = 16

        class nA0_24(Beam):
            _beam_diameter: float = 5.4e-08
            _current: float = 2.4e-10
            _voltage: float = 16

        class nA0_47(Beam):
            _beam_diameter: float = 6.6e-08
            _current: float = 4.7e-10
            _voltage: float = 16

        class nA1_4(Beam):
            _beam_diameter: float = 9.3e-08
            _current: float = 1.4e-09
            _voltage: float = 16

        class nA3_3(Beam):
            _beam_diameter: float = 1.3e-07
            _current: float = 3.3e-09
            _voltage: float = 16

        class nA4_7(Beam):
            _beam_diameter: float = 1.55e-07
            _current: float = 4.7e-09
            _voltage: float = 16

        class nA21(Beam):
            _beam_diameter: float = 5.2e-07
            _current: float = 2.1e-08
            _voltage: float = 16

    # endregion 16kV

    # region 30kV
    class Acceleration30kV:
        class pA1_5(Beam):
            _beam_diameter: float = 7e-09
            _current: float = 1.5e-12
            _voltage: float = 30

        class pA9_7(Beam):
            _beam_diameter: float = 1.3e-08
            _current: float = 9.7e-12
            _voltage: float = 30

        class pA28(Beam):
            _beam_diameter: float = 1.7e-08
            _current: float = 2.8e-11
            _voltage: float = 30

        class pA48(Beam):
            _beam_diameter: float = 1.9e-08
            _current: float = 4.8e-11
            _voltage: float = 30

        class pA93(Beam):
            _beam_diameter: float = 2.4e-08
            _current: float = 9.3e-11
            _voltage: float = 30

        class nA0_28(Beam):
            _beam_diameter: float = 3.1e-08
            _current: float = 2.8e-10
            _voltage: float = 30

        class nA0_46(Beam):
            _beam_diameter: float = 3.5e-08
            _current: float = 4.6e-10
            _voltage: float = 30

        class nA0_92(Beam):
            _beam_diameter: float = 4.4e-08
            _current: float = 9.2e-10
            _voltage: float = 30

        class nA2_8(Beam):
            _beam_diameter: float = 6.6e-08
            _current: float = 2.8e-09
            _voltage: float = 30

        class nA6_5(Beam):
            _beam_diameter: float = 1.02e-07
            _current: float = 6.5e-09
            _voltage: float = 30

        class nA9_3(Beam):
            _beam_diameter: float = 1.29e-07
            _current: float = 9.3e-09
            _voltage: float = 30

        class nA21(Beam):
            _beam_diameter: float = 2.74e-07
            _current: float = 2.1e-08
            _voltage: float = 30
    # endregion 30kV



from src.deFIBrillator import FIBG2

# Initialize the G2 FIB
fib = FIBG2()

# Set the Ion beam. Beams are found under fib.Beams (default beam is 48 pA at 30kV).
fib.set_beam(fib.Beams.Acceleration30kV.pA48)

# Set the magnification of the fib. This can be done by setting it directly, or settings the field of view.
fib.set_fov(5, "um")
# or ie. fib.set_magnification(5000)

# You can print the fib instance to see all specific and derived parameters
print(fib)

# Create a new pattern and specify dwell time and depth
pat = fib.new_pattern()
pat.set_dwell_time(1, "us")
pat.set_depth(100, "nm")

# Create a circle shape
circle = pat.add.circle(radius=100, millable=True)

# Create a 3x3 square lattice with periodicity 300 nm in x and y. Add the circle so it's placed at each lattice point.
# Alpha is the lattice vector in x, and beta in y. Gamma is the angle between them.
lat = pat.add.lattice(rows=3, cols=3, alpha=300, beta=300, gamma=90, shape=circle)

# Plot the pattern
pat.plot_pattern()

# Plot the beam path
pat.plot_beam_path()

# Show animation of the beam path
pat.animate_beam_path()

# Create stream file of the pattern
pat.write_stream("circle_array")

# Optionally you can plot the stream file you created to verify that it's correct. For very low FIB magnifications,
# converting physical coordinates to pixel values can really mess up the beam path, so it's good to check.
from src.deFIBrillator import plot_stream_file
plot_stream_file("circle_array.str")



#
# # Build spatial index
#     tree = STRtree(points)
#
#     # Calculate dwell times
#     for spot in points:
#         # Find all spots within the Euclidian overlap distance
#         neighbors: NDArray = tree.query(buffer(spot, beam_radius), predicate="dwithin", distance=beam_radius)
#         print(neighbors)
#         if neighbors.shape[0] != 0:
#             n = [tree.geometries[int(neighbor)] for neighbor in neighbors]
#             distances = [spot.distance(ne) for ne in n]
#             dose_pr_area_goal = beam.get_dose(mill.get_dwell_time("us") / beam.get_area("nm"))
#             dose_pr_area_actual = beam.get_dose(float(point_dict[spot][0]) / beam.get_area("nm"))
#         else:
#             continue
#
#         dose_goal = dose_pr_area_goal * beam.get_area("nm")  # The dose this spot should ideally have.
#         total_dose = dose_pr_area_actual * beam.get_area("nm")  # Dose contribution from the spot itself.
#
#         this_spot = Point().buffer(beam_radius)
#         for neighbor, distance in zip(neighbors, distances):
#             print(distance)
#             n_spot = Point((distance, 0)).buffer(beam_radius)
#             intersection = this_spot.intersection(n_spot)
#             plt.figure()
#             plt.plot(*intersection.exterior.xy)
#             plt.show()
#             dose = (intersection.area
#                     * beam.get_dose(float(point_dict[tree.geometries[neighbor]][0] * 100))
#                     / beam.get_area("nm")
#                     )
#             total_dose += dose
#
#         print(dose_goal, total_dose, dose_goal/total_dose)
#
#         point_dict[spot][0] *= dose_goal/total_dose
#
#     plt.figure(figsize=(15, 15))
#     for c in streamlines:
#         xx = []
#         yy = []
#         for _, x, y, _ in c:
#             xx.append(x)
#             yy.append(y)
#         plt.plot(xx, yy, "o-")
#     plt.show()
#
#     from scipy.interpolate import griddata
#
#     # Extract x, y, and dose values
#     x = np.array([p.x for p in point_dict.keys()])
#     y = np.array([p.y for p in point_dict.keys()])
#     dose = np.array([v[0] for _, v in point_dict.items()])
#
#     # Define grid resolution
#     grid_x, grid_y = np.mgrid[min(x):max(x):1000j, min(y):max(y):1000j]
#
#     # Interpolate dose values onto the grid
#     grid_z = griddata((x, y), dose, (grid_x, grid_y), method='cubic')
#
#     # Plot the heatmap
#     plt.figure(figsize=(6, 5))
#     plt.imshow(grid_z, extent=(min(x), max(x), min(y), max(y)), origin='lower', cmap='hot')
#     plt.colorbar(label="Dose Value")
#     plt.scatter(x, y, c='blue', marker='o', label="Data Points")  # Original points
#     plt.legend()
#     plt.title("Dose Heatmap")
#     plt.xlabel("X Coordinate")
#     plt.ylabel("Y Coordinate")
#     plt.show()
#
#     return streamlines


# Std lib
from abc import ABC
from typing import TypeVar

# 3rd party lib
import numpy as np

# Local lib
from ..resources import (length_units, current_units, voltage_units, LENGTHUNITS, CURRENTUNITS, VOLTAGEUNITS,
                         TIMEUNITS, time_units as t_u, DOSEUNITS, dose_units as d_u)


BeamType = TypeVar("BeamType", bound="Beam")


class Beam(ABC):

    # region Class Body

    _beam_diameter: float
    _current: float
    _voltage: float

    # endregion Class Body

    # region Dev Methods

    def __repr__(self) -> str:
        if self.get_current() < 100:
            current = f"{self.get_current("pa")} pA"
        else:
            current = f"{self.get_current("na")} nA"
        return (f"Ion Beam:\n"
                f"\tAcceleration voltage: {self.get_voltage("kv")} V\n"
                f"\tCurrent: {current}\n"
                f"\tBeam Diameter: {self.get_diameter("nm")} nm\n"
                f"\tBeam Radius: {self.get_radius("nm")} nm\n"
                f"\tBeam Area: {self.get_area("nm")} nm²")

    # endregion Dev Methods

    # region User Methods

    def get_dose(self, dwell_time: float, time_units: TIMEUNITS = "us", dose_units: DOSEUNITS = "nc") -> float:
        """Calculates and returns the dose in the specified dose units provided to a beam spot in a
        given dwell time [specified time units]."""
        return d_u((self._current * t_u(dwell_time, time_units, "s")) / ((self.get_radius("um")) ** 2),
                   "c", dose_units)

    def get_current(self, units: CURRENTUNITS = "na") -> float:
        """Return the beam current in the provided current units."""
        return current_units(self._current, "a", units)

    def get_area(self, units: LENGTHUNITS = "nm") -> float:
        """Return the beam area in the desired units squared."""
        return np.pi * np.pow(self.get_radius(units), 2)

    def get_diameter(self, units: LENGTHUNITS = "nm") -> float:
        """Returns the beam diameter in the desired length units.."""
        return length_units(self._beam_diameter, "m", units)

    def get_radius(self, units: LENGTHUNITS = "nm") -> float:
        """Returns the beam radius in the desired length units."""
        return self.get_diameter(units) / 2

    def get_voltage(self, units: VOLTAGEUNITS) -> float:
        """Returns the acceleration voltage of the beam in the desired voltage units."""
        return voltage_units(self._voltage, "v", units)

    # endregion User Methods




from .rectangle import Rectangle, RectangleKwargs
from .regular_polygon import RegularPolygon, RegularPolygonKwargs
from .polygon import Poly, PolyKwargs
from .circle import Circle, CircleKwargs
from .shape import Shape, CombinedShape
from .lattice import Lattice, LatticeKwargs

__all__ = ["Rectangle", "RectangleKwargs", "RegularPolygon", "RegularPolygonKwargs", "Poly", "PolyKwargs",
           "Circle", "CircleKwargs", "Shape", "CombinedShape", "Lattice", "LatticeKwargs"]



from __future__ import annotations

# Std lib
from typing import Unpack, TypedDict

# 3rd party lib
from matplotlib.font_manager import FontProperties
from matplotlib.textpath import TextPath
from shapely import Polygon, MultiPolygon, unary_union

# Local lib
from .shape import Shape


class TextKwargs(TypedDict, total=False):
    x: float
    y: float
    font_size: int
    millable: bool


class Text(Shape):

    # region Class Body

    # Class variables
    _default_geometry = Shape._default_geometry.copy()
    _default_geometry.update({"fontsize": 1000})

    # Instance variables
    _fontsize: int
    _text: str

    __slots__ = ["_fontsize", "_text"]

    # endregion Class Body

    # region Dev Methods

    def __init__(self, text: str, **kwargs: Unpack[TextKwargs]) -> None:
        super().__init__(**kwargs)
        self._text = text

    def _get_polygon(self) -> MultiPolygon:
        """
        Converts a given text string into a Shapely MultiPolygon.

        Parameters:
        text (str): The text string to convert.
        font_path (str, optional): Path to a .ttf font file. Defaults to the system's default font.
        size (int, optional): Font size. Defaults to 1.

        Returns:
        MultiPolygon: The resulting MultiPolygon geometry of the text.
        """
        # Set up font properties
        font_prop = FontProperties()

        # Create the text path at origin (0, 0)
        path = TextPath(self._get_absolute_position(), self._text, size=self._fontsize * (1 / 0.725), prop=font_prop)

        # Extract polygons from the path
        polygons = []
        for vertices in path.to_polygons():
            if len(vertices) >= 3:  # Ensure polygon has at least 3 points
                polygons.append(Polygon(vertices))

        # Union polygons to handle overlaps and holes
        if not polygons:
            return MultiPolygon()
        union = unary_union(polygons)

        # Return as a MultiPolygon
        if union.geom_type == 'Polygon':
            return MultiPolygon([union])
        return union

    # endregion Dev Methods

    # region User Methods

    def copy(self, **kwargs) -> Text:
        return super().copy(**kwargs)

    # endregion User Methods

    # region Properties

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, text: str):
        self._text = str(text)

    @property
    def font_size(self) -> int:
        return self._fontsize

    @font_size.setter
    def font_size(self, value: int):
        self._fontsize = int(value)

    # endregion Properties



from __future__ import annotations

# Std lib
import math
from abc import abstractmethod
from copy import copy
from typing import Tuple, ClassVar, Dict, Any, TypedDict, Unpack, List, Union, Literal, Iterable

# 3rd party lib
from matplotlib import pyplot as plt
from shapely import unary_union
from shapely.affinity import translate, rotate
from shapely.geometry import Polygon, MultiPolygon, Point


# region Constants and Literals

BOUNDARIES = Literal["x_min", "x_max", "y_min", "y_max"]


# endregion Constants and Literals


# region Shape Base Class
class Shape:
    # region Class Body

    # Class variables
    _default_geometry: ClassVar[Dict] = {"x": 0, "y": 0, "rotation": 0, "millable": True}

    # Instance variables
    _x: float
    _y: float
    _rotation: float
    _millable: bool
    _pattern: Any

    __slots__ = ["_x", "_y", "_rotation", "_millable", "_pattern"]

    # endregion Class Body

    # region Dev Methods

    def __init__(self, pattern=None, **kwargs) -> None:

        if pattern is not None:
            self._pattern = pattern

        # Use the default values for not passed arguments
        default = self._default_geometry.copy()
        default.update(kwargs)
        for k, v in default.items():
            setattr(self, "_" + k, v)

    def _apply_translation(self, polygon: Polygon) -> Polygon:
        shape = translate(polygon, *self._get_absolute_position())
        return shape

    def _apply_rotation(self, polygon: Polygon) -> Polygon:
        return rotate(polygon, self._rotation)

    def _get_min(self) -> Tuple[float, float]:
        """Returns a tuple of the minimum x and y coordinates."""
        x, y = self.polygon.exterior.xy
        return min(x), min(y)

    def _get_max(self) -> Tuple[float, float]:
        """Returns a tuple of the maximum x and y coordinates."""
        x, y = self.polygon.exterior.xy
        return max(x), max(y)

    def _add_to_pattern(self) -> None:
        """Adds the shape to the provided pattern."""
        if self._pattern is not None:
            self._pattern._add_shape(self)

    def _get_absolute_position(self) -> Tuple[float, float]:
        """Returns the absolute position of the shape, meaning the position of the parent pattern plus the
        position of the shape."""
        if self._pattern is None:
            return self._x, self._y
        else:
            return self._x + self._pattern.x, self._y + self._pattern.y

    @abstractmethod
    def _get_polygon(self) -> Polygon | MultiPolygon:
        ...

    # endregion Dev Methods

    # region User Methods

    def plot(self) -> None:
        plt.figure(figsize=(15, 10))
        polygon = self.polygon
        if polygon.geom_type == "MultiPolygon":
            for poly in polygon.geoms:
                if self._millable:
                    plt.fill(*poly.exterior.xy)
                else:
                    plt.plot(*poly.exterior.xy)
        else:
            if self._millable:
                plt.fill(*polygon.exterior.xy)
            else:
                plt.plot(*polygon.exterior.xy)
        plt.gca().set_aspect("equal")
        plt.show()

    def set_position(self, x: float | Shape = None, y: float | Shape = None) -> None:
        """
        Set position using cartesian coordinates.

        Args:
            x (float | Shape): The new x-coordinate. Uses previous coordinate if not provided.
            y (float): The new y-coordinate. Uses previous coordinate if not provided.
        """
        self._x = x if x is not None else self._x
        self._y = y if y is not None else self._y

    def set_position_polar(self,
                           radius: float,
                           theta: float,
                           origin: tuple[float, float] | Shape = (0.0, 0.0)) -> None:
        """
        Set position using polar coordinates relative to a specified origin

        Args:
            radius: Distance from origin to new position
            theta: Angle in degrees (counterclockwise from positive x-axis)
            origin: Tuple (x, y) specifying the reference point. Can be another shape, then it's position is used.
                    Defaults to (0, 0) if not provided
        """

        # Unpack origin coordinates
        if isinstance(origin, Shape):
            origin_x, origin_y = origin.x, origin.y
        else:
            origin_x, origin_y = origin

        # Convert angle to radians
        theta_rad = math.radians(theta)

        # Convert polar to Cartesian coordinates relative to origin
        x = origin_x + radius * math.cos(theta_rad)
        y = origin_y + radius * math.sin(theta_rad)

        # Update position using existing Cartesian method
        self.set_position(x=x, y=y)

    def set_rotation(self, rot_angle: float) -> None:
        """Sets the counterclockwise rotation of the shape."""
        self._rotation = float(rot_angle)

    def rotate(self, rot_angle: float, rot_point: Tuple[float, float] | Shape = None) -> rotate:
        """Rotates the shape counterclockwise from it's current rotation around a point. If no rotation point is
        provided, the shape is rotated around it's center position. Another shape can be provided as the rotation point.
        The method can be stacked, ie. shape.rotate(45)(36)."""

        self._rotation = self._rotation + rot_angle
        if isinstance(rot_point, Shape):
            rot_point = (rot_point.x, rot_point.y)
        elif rot_point is None:
            rot_point = (self._x, self._y)
        elif not isinstance(rot_point, (tuple, list)):
            raise TypeError(f"Expected iterable type, got '{type(rot_point)}'.")

        point = rotate(Point(self._x, self._y), self._rotation, rot_point)
        self.set_position(point.x, point.y)

        return self.rotate

    def place_next_to(self, boundary: BOUNDARIES, shape: Shape, offset: float = 0) -> None:
        """Places the shape adjacent to the specified boundary of the provided shape."""
        axis = boundary[0]
        scale = -1 if boundary[2:] == "min" else 1
        coordinate = getattr(shape, boundary)
        setattr(self, boundary[0],
                coordinate
                + offset
                + scale
                * ((getattr(self, axis + "_max") - getattr(self, axis + "_min")) / 2)
                )

    @abstractmethod
    def copy(self, **kwargs):
        """Returns an exact copy of the shape, modified according to the provided kwargs."""

        # 'Secret' keyword for not adding copied shape to parent pattern
        add = kwargs.pop("add_to_pattern", True)

        new = copy(self)

        for k, v in kwargs.items():
            setattr(new, "_" + k, v)

        if add:
            new._add_to_pattern()

        return new

    def combine(self, other: Union[Iterable[Union[Shape]], Shape]) -> CombinedShape:
        """
        Combines the shape with other shapes to form a composite shape.
        """

        # Remove shapes from the pattern
        self._pattern._shapes.remove(self)
        if isinstance(other, (list, tuple)):
            for o in other:
                if o in self._pattern._shapes:
                    self._pattern._shapes.remove(o)
        else:
            self._pattern._shapes.remove(other)

        # Extract shapes
        shapes = []

        def extract_shapes(shape: Shape) -> None:
            if isinstance(shape, CombinedShape):
                for subshape in shape._shapes:
                    extract_shapes(subshape)
            elif isinstance(shape, Shape):
                shapes.append(shape)
            elif isinstance(shape, (list, tuple)):
                for s in shape:
                    extract_shapes(s)
            else:
                raise TypeError(f"Expected type 'Shape' or subclass, got '{type(shape)}'.")

        extract_shapes(self)
        extract_shapes(other)

        # Create combined shape and add it to pattern
        combined = CombinedShape(self._pattern)
        combined._shapes = shapes
        combined._add_to_pattern()

        return combined

    # endregion User Methods

    # region Properties

    @property
    def polygon(self) -> Polygon | MultiPolygon:
        """Returns the polygon associated with the shape."""
        return self._get_polygon()

    @property
    def x(self) -> float:
        return self._x

    @x.setter
    def x(self, value: float):
        self._x = float(value)

    @property
    def y(self) -> float:
        return self._y

    @y.setter
    def y(self, value: float):
        self._y = float(value)

    @property
    def x_min(self) -> float:
        return self._get_min()[0]

    @property
    def x_max(self) -> float:
        return self._get_max()[0]

    @property
    def y_min(self) -> float:
        return self._get_min()[1]

    @property
    def y_max(self) -> float:
        return self._get_max()[1]

    @property
    def rotation(self) -> float:
        return self._rotation

    @rotation.setter
    def rotation(self, value: float):
        self.set_rotation(value)

    @property
    def area(self) -> float:
        return self.polygon.area

    @property
    def millable(self) -> bool:
        return self._millable

    @millable.setter
    def millable(self, millable: bool) -> None:
        """Defines wether the shape's interior should be milled or not."""
        if not isinstance(millable, bool):
            raise TypeError(f"Expected 'bool', got '{type(millable)}'.")
        self._millable = millable

    # endregion Properties


# endregion Shape Base Class


# region Combined Shape
class CopyKwargs(TypedDict, total=False):
    x: float
    y: float
    rotation: float


class CombinedShape(Shape):
    # region Class Body

    # Class variables
    _default_geometry = Shape._default_geometry.copy()

    # Instance variables
    _shapes: List[Shape]

    # endregion Class Body

    # region Dev Methods

    def _get_polygon(self) -> Polygon:

        millable = []
        non_millable = []
        for shape in self._shapes:
            if shape._millable:
                millable.append(shape.polygon)
            else:
                non_millable.append(shape.polygon)

        total_millable = unary_union(millable)
        total_non_millable = unary_union(non_millable)

        polygon = total_millable.difference(total_non_millable)

        polygon = self._apply_translation(polygon)
        polygon = self._apply_rotation(polygon)
        return polygon

    # endregion Dev Methods

    # region User Methods

    def rotate(self, rot_angle: float, rot_point: Tuple[float, float] | Shape = None) -> rotate:
        for shape in self._shapes:
            shape.rotate(shape.rotation - self._rotation + rot_angle, rot_point=self)
        super().rotate(rot_angle, rot_point)

    def set_rotation(self, rot_angle: float) -> None:
        for shape in self._shapes:
            rotate_by = shape.rotation - self._rotation + rot_angle
            shape.rotate(rotate_by, rot_point=self)

    def set_position(self, x: float | Shape = None, y: float | Shape = None) -> None:
        for shape in self._shapes:
            shape.x = shape.x - self._x + x if x is not None else shape.x
            shape.y = shape.y - self._y + y if y is not None else shape.y
        super().set_position(x, y)

    def copy(self, **kwargs: Unpack[CopyKwargs]) -> CombinedShape:
        new_shapes = [shape.copy(add_to_pattern=False) for shape in self._shapes]
        new_combined = CombinedShape(self._pattern)
        new_combined._shapes = new_shapes
        for k, v in kwargs.items():
            setattr(new_combined, k, v)
        new_combined._add_to_pattern()
        return new_combined

    # endregion User Methods

    # region Properties

    @property
    def x(self) -> float:
        return self._x

    @x.setter
    def x(self, x: float) -> None:
        for shape in self._shapes:
            shape.x = shape.x - self._x + x
        self._x = x

    @property
    def y(self) -> float:
        return self._y

    @y.setter
    def y(self, y: float) -> None:
        for shape in self._shapes:
            shape.y = shape.y - self._y + y
        self._y = y

    @property
    def x_min(self) -> float:
        return min([shape.x_min for shape in self._shapes]) + self._x

    @property
    def x_max(self) -> float:
        return max([shape.x_max for shape in self._shapes]) + self._x

    @property
    def y_min(self) -> float:
        return min([shape.y_min for shape in self._shapes]) + self._y

    @property
    def y_max(self) -> float:
        return max([shape.y_max for shape in self._shapes]) + self._y

    @property
    def x_span(self) -> float:
        return self.x_max - self.x_min

    @property
    def y_span(self) -> float:
        return self.y_max - self.y_min

    # endregion Properties

# endregion Combined Shape



# Std lib
from abc import ABC, abstractmethod
from typing import ClassVar, List, Type

# 3rd party lib
import numpy as np
from matplotlib import pyplot as plt

# Local lib
from .resources import MagnificationHint
from ..beams import BeamType
from ..pattern import Pattern
from ..resources import LENGTHUNITS, length_units


class FIB(ABC):

    # region Class Body

    # Class variables
    _fov_1000x: ClassVar[float]
    _fov_pixels: ClassVar[int]
    _default_beam: Type[BeamType]

    # Instance variables
    _magnification: int
    _beam: BeamType
    _patterns: List[Pattern]

    # endregion Class Body

    # region Dev Methods

    def __init__(self, magnification: int = None) -> None:
        if magnification is None:
            self._magnification = 1000  # Default magnification is 1000x
        else:
            self.set_magnification(magnification)
        self._patterns = []
        self._beam = self._default_beam()

    # endregion Dev Methods

    # region User Methods

    def set_magnification(self, magnification: int) -> None:
        """Sets the magnification of the FIB instance."""
        self._magnification = magnification

    def get_fov(self, units: LENGTHUNITS = "nm") -> float:
        """Gets the field of view of the FIB monitor in nanometers based on the current magnification."""
        return length_units(self._fov_1000x, "nm", units) * (1000 / self._magnification)

    def set_fov(self, fov: float, units: LENGTHUNITS = "nm"):
        """Sets the magnification that provides the field of view closest to the one provided. Returns a tuple
        with the magnification, and the actual field of view in the same units as provided."""
        closest_mag = int(np.round((1000 / length_units(fov, units, "nm")) * self._fov_1000x))
        fov = (1000 / closest_mag) * self._fov_1000x
        self._magnification = closest_mag
        return closest_mag, length_units(fov, "nm", units)

    def get_pixelsize(self, units: LENGTHUNITS = "nm") -> float:
        """Gets the size of each pixel in the field of view in nanometers [nm]."""
        return self.get_fov(units) / self._fov_pixels

    def new_pattern(self, x: float = 0, y: float = 0) -> Pattern:
        pattern = Pattern(self, x, y)
        self._patterns.append(pattern)
        return pattern

    def set_beam(self, beam: Type[BeamType]) -> None:
        self._beam = beam()

    def get_magnification_hint(self, desired_pixelsize: float, units: LENGTHUNITS = "nm") -> MagnificationHint:
        closest_mag = int(np.round((1000 * self._fov_1000x
                                    / (length_units(desired_pixelsize, units, "nm") * self._fov_pixels))))
        effective_pixelsize = (1000 * self._fov_1000x) / (closest_mag * self._fov_pixels)
        fov = (1000 / closest_mag) * self._fov_1000x
        return closest_mag, effective_pixelsize, length_units(fov, "nm", units)

    def get_magnification_stats(self, magnification: int, units: LENGTHUNITS = "nm") -> MagnificationHint:
        effective_pixelsize = (1000 * self._fov_1000x) / (magnification * self._fov_pixels)
        fov = (1000 / magnification) * self._fov_1000x
        return magnification, effective_pixelsize, length_units(fov, "nm", units)

    def draw_patterns(self) -> None:
        ax = None
        for pattern in self._patterns:
            ax = pattern._get_pattern_ax(ax)
        ax.relim()
        ax.autoscale_view()
        ax.set_aspect('equal')
        plt.show()

    # endregion User Methods

    # region Properties

    @property
    def beam(self) -> BeamType:
        return self._beam

    @property
    def magnification(self) -> int:
        """Gets the magnification of the FIBG2 instance."""
        return self._magnification

    @magnification.setter
    def magnification(self, magnification: int) -> None:
        """Sets the magnification of the FIBG2 instance."""
        if not isinstance(magnification, int):
            raise TypeError(f"Got '{type(magnification)}', expected 'type(int)'.")
        elif magnification < 150:
            raise ValueError(f"Magnification '{magnification}' is lower than the FIB's lowest magnification of 150x.")
        else:
            self._magnification = magnification

    # endregion Properties



# Local lib
import sys

from .fib import FIB
from ..beams import IonBeams


class FIBG2(FIB):

    # region Class Body

    # Class variables
    _fov_1000x = 128_000
    _fov_pixels = 65536
    _default_beam = IonBeams.Acceleration30kV.pA48

    # Instance variables
    Beams: IonBeams

    # endregion Class Body

    # region Dev Methods

    def __init__(self, magnification: int = None) -> None:
        super().__init__(magnification)
        self.Beams = IonBeams()

    def __repr__(self) -> str:
        return (f"FEI Helios Nanolab Dualbeam G2:\n"
                f"\tMagnification: {self._magnification}\n"
                f"\tField of view: {self.get_fov("um")} um\n"
                f"\tPixelsize: {self.get_pixelsize("nm")} nm\n"
                f"{self.beam}")

    # endregion Dev Methods



import matplotlib.pyplot as plt


def plot_stream_file(file_path: str, title: str = None) -> None:

    if not file_path.endswith(".str"):
        file_path += ".str"

    segments = []  # List to store segments (each segment is a list of (x, y) tuples)
    current_segment = []  # List to store the current segment

    # Open the file and read all lines.
    with open(file_path, 'r') as f:
        lines = f.readlines()

    # Skip the first three header lines.
    for line in lines[3:]:
        line = line.strip()
        if not line:  # Skip empty lines, if any.
            continue

        # Split the line into parts.
        parts = line.split()

        # We expect at least three numbers per line.
        if len(parts) < 3:
            continue

        # The second and third elements are x and y pixel coordinates.
        # (Remember: list indices start at 0, so parts[1] is the second element.)
        x = float(parts[1])
        y = float(parts[2])

        # Check if there is a fourth element.
        if len(parts) >= 4:
            if int(parts[3]) == 0:
                current_segment.append((x, y))
                segments.append(current_segment.copy())
                current_segment = []
            elif int(parts[3]) == 1:
                current_segment.append((x, y))
        else:
            current_segment.append((x, y))

    # Now, plot each segment.
    for seg in segments:
        # Unzip the list of (x,y) tuples into separate x and y lists.
        xs, ys = zip(*seg)
        plt.plot(xs, ys, marker='o')  # You can remove or adjust marker style if desired.

    plt.xlabel("X Pixels")
    plt.ylabel("Y Pixels")
    plt.title("Streamlines")
    plt.gca().set_aspect("equal")
    plt.show()



from __future__ import annotations

# Std lib
import math
from typing import TypedDict, Unpack

# 3rd party lib
from shapely import Polygon

# Local lib
from .shape import Shape


class RegularPolygonKwargs(TypedDict, total=False):
    millable: bool
    nr_sides: int
    x: float
    y: float
    inner_radius: float
    outer_radius: float
    side_length: float
    rotation: float


class RegularPolygon(Shape):

    # region Class Body

    # Class variables
    _default_geometry = Shape._default_geometry.copy()
    _default_geometry.update({"nr_sides": 6, "outer_radius": 50.})

    # Instance variables
    _nr_sides: int
    _outer_radius: float

    # endregion Class Body

    # region Dev Methods

    def __init__(self, **kwargs: Unpack[RegularPolygonKwargs]):

        nr_sides = kwargs.pop("nr_sides", None)
        if nr_sides is not None:
            if nr_sides < 3:
                raise ValueError(f"The numer of sides of a regular polygon must be greater than 2, got '{nr_sides}'.")
            else:
                self._nr_sides = nr_sides
        else:
            self._nr_sides = self._default_geometry["nr_sides"]

        kwargs["outer_radius"] = self._to_outer_radius(**kwargs)

        super().__init__(**kwargs)

    def _get_polygon(self, **kwargs: Unpack[RegularPolygonKwargs]) -> Polygon:

        # Generate regular polygon vertices
        angles = [90 + i * (360 / self.nr_sides) for i in range(self.nr_sides)]
        points = [
            (self._outer_radius * math.cos(math.radians(angle)),
             self._outer_radius * math.sin(math.radians(angle)))
            for angle in angles
        ]
        polygon = Polygon(points)

        # Apply translation and rotation if needed
        polygon = self._apply_translation(polygon)
        polygon = self._apply_rotation(polygon)

        # Initialize the shape with final polygon
        return polygon

    def _to_outer_radius(self, **kwargs) -> float:

        inner_radius = kwargs.pop("inner_radius", None)
        outer_radius = kwargs.pop("outer_radius", None)
        side_length = kwargs.pop("side_length", None)

        # Validate only one size parameter is provided
        if len([a for a in [inner_radius, outer_radius, side_length] if a is not None]) > 1:
            raise ValueError("Expected either 'inner_radius', 'outer_radius', or 'side_length', got multiple.")

        if inner_radius is not None:
            return inner_radius / math.cos(math.pi / self._nr_sides)  # type: ignore
        elif side_length is not None:
            return side_length / (2 * math.sin(math.pi / self._nr_sides))  # type: ignore
        elif outer_radius is not None:
            return outer_radius  # type: ignore

    # endregion Dev Methods

    # region User Methods

    def copy(self, **kwargs: Unpack[RegularPolygonKwargs]) -> RegularPolygon:
        return super().copy(**kwargs)

    # endregion User Methods

    # region Properties

    @property
    def nr_sides(self) -> int:
        """Returns the polygon's number of sides."""
        return self._nr_sides

    @nr_sides.setter
    def nr_sides(self, nr: int) -> None:
        """Sets the polygon's number of sides."""
        self._nr_sides = nr

    @property
    def side_length(self) -> float:
        """Returns the side lenght of the polygon."""
        return self._outer_radius * (2 * math.sin(math.pi / self._nr_sides))

    @side_length.setter
    def side_length(self, length: float) -> None:
        """Sets the side length og the polygon."""
        self._outer_radius = self._to_outer_radius(side_length=length)

    @property
    def inner_radius(self) -> float:
        """Returns the inner radius of the polygon."""
        return self._outer_radius * math.cos(math.pi / self._nr_sides)

    @inner_radius.setter
    def inner_radius(self, radius: float) -> None:
        """Sets the inner radius of the polygon."""
        self._to_outer_radius(inner_radius=radius)

    @property
    def outer_radius(self) -> float:
        return self._outer_radius

    @outer_radius.setter
    def outer_radius(self, radius: float) -> None:
        """Sets the outer radius of the polygon."""
        self._outer_radius = radius

    # endregion Properties



from typing import Tuple, Literal

MagnificationHint = Tuple[int, float, float]
ScanStyles = Literal["contour inwards", "contour outwards", "serpentine single horizontal",
                     "serpentine single vertical", "serpentine double horizontal", "serpentine double vertical",
                     "raster left-right bottom-top", "raster right-left, bottom top", "raster left-right top-bottom",
                     "raster right-left top-bottom"]
Voltages = Literal["500V", "1kV", "2kV", "5kV", "8kV", "16kV", "30kV"]


# Std lib
from typing import List

# 3rd party lib
import numpy as np
from numpy import array
from numpy.typing import NDArray
from shapely import buffer, Polygon, MultiPolygon, LineString

# Local lib
from ..beams import BeamType
from ..mill import Mill


def contour(shape: Polygon | MultiPolygon, mill: Mill, beam: BeamType, rasterize: bool = True) -> List[NDArray]:
    """Generates a scan path consisting of shrinkingcontours from edges to center."""

    beam_radius = beam.get_radius(units="nm")
    overlap = mill.get_overlap()
    step = 2 * beam_radius * (1 - overlap)

    contours = []

    def add_linestrings(geometry):
        if isinstance(geometry, Polygon):
            if not geometry.exterior.is_empty:
                contours.append(geometry.exterior)
            for interior in geometry.interiors:
                contours.append(interior)
        elif isinstance(geometry, MultiPolygon):
            for polygon in geometry.geoms:
                if not polygon.exterior.is_empty:
                    contours.append(polygon.exterior)
                for interior in polygon.interiors:
                    contours.append(interior)

    k = 0
    buffered = buffer(shape, -beam_radius, join_style="mitre")

    while True:

        if k != 0:
            buffered: Polygon = buffer(buffered, -step, join_style="round", quad_segs=100)

            if buffered.is_empty:
                break

        add_linestrings(buffered)
        k += 1

    # Extract all spots from contours
    streamlines = []
    if rasterize:

        for line in contours:
            line: LineString
            first_spot = True
            cont = []
            nr_steps = max(2, round(line.length / step))
            if nr_steps == 2:
                points: List[Polygon] = [line.centroid]
            else:
                points: List[Polygon] = []
                for distance in np.linspace(0, line.length, nr_steps):
                    points.append(line.interpolate(distance))

            if len(points) > 1:
                _idx = slice(None, -1, None)
            else:
                _idx = slice(None, None, None)

            for point in points[_idx]:
                if first_spot:
                    spot_line = array([int(mill.get_dwell_time("ns") / 100), point.centroid.x, point.centroid.y, 1])
                    cont.append(spot_line)
                    first_spot = False
                else:
                    spot_line = array([int(mill.get_dwell_time("ns") / 100), point.centroid.x, point.centroid.y, 2])
                    cont.append(spot_line)

            if line.is_closed:
                cont.append(array([1, cont[0][1], cont[0][2], 0]))
            else:
                cont.append(array([1, cont[-1][1], cont[-1][2], 0]))
            streamlines.append(array(cont))

    else:
        for line in contours:
            c = []
            for coordinate in line.coords:
                c.append(array([0, coordinate[0], coordinate[1], 0]))
            streamlines.append(array(c))

    return streamlines
